"""
Скрипт для перехвата DNS-трафика, проходящего
через сетевую карту по умолчанию. А также детектирования
обращений к IP-адресам Яндекс и Google, чтобы проверить,
насколько часто происходит к ним обращение.
При детекте выводиться звуковой сигнал. Также проверяется
принадлежность ip-адреса домена к Cloudflare.
Для работы скрипта необходимо установить следующие библиотеки:
pip install chime colorama requests getmac
pip install --pre scapy[basic]
"""
import os.path
import platform
import subprocess
from datetime import datetime
from ipaddress import ip_network, ip_address
from socket import gethostbyname, gaierror, socket, AF_INET, SOCK_DGRAM

import chime
import getmac
from colorama import Fore
from colorama import init
from requests import get
from scapy.all import sniff

init()

list_addr = []
yandex = []
google = []


def local_ipv4():
    """
    Выполняется запрос на адрес.
    После этого забирается адрес сокета.
    :return: локальный ip-адрес.
    """
    st = socket(AF_INET, SOCK_DGRAM)
    try:
        st.connect(('10.255.255.255', 1))
        ip_local = st.getsockname()[0]
    except Exception:
        ip_local = '127.0.0.1'
    finally:
        st.close()
    return ip_local


def ip_iface():
    """
    Получаем имя активного сетевого интерфейса в Linux.
    :return: имя активного сетевого интерфейса.
    """
    return subprocess.check_output('ip -h -br a | grep UP', shell=True).decode().split()[0].strip()


def iface_win():
    """
    Получаем имя активного сетевого интерфейса в Windows.
    :return: имя активного сетевого интерфейса.
    """
    interface_all = subprocess.check_output('getmac /FO csv /NH /V', shell=False).decode('cp866').splitlines()
    for line in interface_all:
        if getmac.getmac.get_mac_address().replace(":", "-").upper() in line:
            return line.split(",")[0].replace('"', '')


def cloud_detect(ip):
    """
    Проверка, входит ли проверяемый ip-адрес
    в диапазон адресов Cloudflare.
    :param ip: ip-адрес для проверки.
    :return: True - найден, False - не найден.
    """
    global list_addr
    for addr in list_addr:
        if ip_address(ip) in ip_network(addr):
            return True
    return False


def change_ip(ip):
    """
    Выполняется проверка, входит ли адрес в
    диапазон адресов Cloudflare. Если да, завершаем
    работу функции. Если нет, проверяем, входит ли
    ip-адрес в диапазоны адресов Яндекс или Google.
    Если да, выводим сообщение в терминал и подаем
    звуковое оповещение.
    :param ip: ip-адрес для проверки.
    :return: выход из функции.
    """
    global yandex
    global google

    if cloud_detect(ip):
        print(Fore.RED + '[-] Обнаружен адрес Cloudflare!')
        chime.theme('mario')
        chime.success()
        return

    for addr_y in yandex:
        if ip_address(ip) in ip_network(addr_y.strip()):
            print(Fore.RED + f'[-] Внимание! Запрос к Yandex: {Fore.RESET}{ip}')
            chime.theme('chime')
            chime.success()
            return
    for addr_g in google:
        if ip_address(ip) in ip_network(addr_g.strip()):
            print(Fore.RED + f'[-] Внимание! Запрос к Google: {Fore.RESET}{ip}')
            chime.theme('big-sur')
            chime.success()
            return


def get_host(domain):
    """
    Попытка получить ip-адрес домена.
    :param domain: доменное имя для получения ip.
    :return: ip-адрес домена или сообщение об ошибке,
    в случае невозможности получить адрес.
    """
    try:
        return gethostbyname(domain)
    except gaierror:
        return 'Unknown - No address associated with hostname'


def packet_sniff(packet):
    """
    Получаем url-сайта из перехваченного пакета.
    Забираем из url доменное имя. Выводим сообщения в
    терминал. Пытаемся получить ip-адрес из DNS Resource Record.
    Если адреса нет, обрабатываем исключение, и пытаемся получить
    ip-адрес по доменному имени.
    В любом из случаев выводим сообщение в терминал об url и
    времени запроса. Запускаем функцию проверки ip-адреса на
    принадлежность к адресам Яндекс, Google или Cloudflare.
    :param packet: перехваченный пакет.
    :return: выход из функции.
    """
    domain = None
    qname = packet['DNS Question Record'].qname.decode()
    try:
        domain = f'{qname[0:-1].split(".")[-2]}.{qname[0:-1].split(".")[-1]}'
    except IndexError:
        print(Fore.GREEN + f'[+] Запрос: {Fore.RESET}{qname[0:-1]}')
        print(Fore.GREEN + f'[+] Время запроса: {Fore.YELLOW}{datetime.now().time()}')
        print(Fore.GREEN + '[+] Домен: Unknown')

    try:
        rdata = packet['DNS Resource Record'].rdata.decode()
        if rdata.split(".")[0] == local_ipv4().split(".")[0]:
            return
        print(Fore.GREEN + f'[+] Запрос: {Fore.RESET}{qname[0:-1]}')
        print(Fore.GREEN + f'[+] Домен: {Fore.YELLOW}{domain}')
        print(Fore.GREEN + f'[+] Время запроса: {Fore.YELLOW}{datetime.now().time()}')
        print(Fore.GREEN + f'[+] IP-адрес домена: {Fore.RESET}{rdata}')
        change_ip(rdata)
        print('\n')
    except IndexError:
        ip = get_host(domain)
        print(Fore.GREEN + f'[+] Запрос: {Fore.RESET}{qname[0:-1]}')
        print(Fore.GREEN + f'[+] Домен: {Fore.YELLOW}{domain}')
        print(Fore.GREEN + f'[+] Время запроса: {Fore.YELLOW}{datetime.now().time()}')
        print(Fore.GREEN + f'[+] IP-адрес домена: {Fore.RESET}{ip}')
        if ip != 'Unknown - No address associated with hostname':
            change_ip(ip)
        print('\n')


def load_range():
    """
    Проверяем, существует ли папка с диапазонами
    адресов. Если нет, выходим из функции. Если есть,
    получаем диапазоны адресов Cloudflare, присваиваем их
    в глобальную переменную list_addr.
    Загружаем диапазоны адресов Яндекс и Google, разбиваем
    построчно и присваиваем в глобальные переменные yandex
    и google.
    :return: выход из функции.
    """
    global list_addr
    global yandex
    global google

    dir_range = os.path.join(os.getcwd(), 'ip_range')
    if not os.path.exists(dir_range):
        print(Fore.RED + '[-] Не существует директории с конфигурацией')
        exit(0)

    print(Fore.GREEN + '\n[*] Получение диапазона адресов Cloudflare')
    list_addr = get(url='https://www.cloudflare.com/ips-v4').text.splitlines()
    list_addr.append("104.16.0.0/12")

    if os.path.exists(os.path.join(dir_range, 'yandex_ipv4.txt')):
        print(Fore.GREEN + '[*] Загрузка адресов Яндекс')
        with open(os.path.join(dir_range, 'yandex_ipv4.txt'), 'r', encoding='utf-8') as ya:
            yandex = ya.read().splitlines()

    if os.path.exists(os.path.join(dir_range, 'google_ipv4.txt')):
        print(Fore.GREEN + '[*] Загрузка адресов Google\n')
        with open(os.path.join(dir_range, 'google_ipv4.txt'), 'r', encoding='utf-8') as go:
            google = go.read().splitlines()


def main():
    """
    Выполняем проверку на операционную систему. В зависимости
    от этого проверяем, если под linux, запущен ли скрипт от
    имени суперпользователя. Если да, запускаем перехват трафика.
    В ОС Windows скрипт не требует наличия административных прав.
    Поэтому перехват трафика запускается сразу.
    Функция перехвата трафика имеет следующие параметры:
    :param iface: сетевой интерфейс для прослушивания.
    :param filter: фильтр, в данном случае фильтруем исходящие запросы на порт 53,
    который является портом DNS.
    :param count: количество перехваченных пакетов. В данном случае, перехват будет
    идти бесконечно.
    :paran store: не сохранять ни один полученный пакет.
    :param prn: функция в которую передаются пакеты для обработки.
    :return: выход из функции.
    """
    if platform.system() == "Linux":
        from os import getuid
        if not getuid() == 0:
            print('\n[+] Запустите скрипт с правами суперпользователя!\n')
            return
        load_range()
        print(Fore.GREEN + '[*] Перехват пакетов...\n')
        sniff(iface=ip_iface(), filter='dst port 53', count=0, store=False, prn=packet_sniff)
    elif platform.system() == "Windows":
        load_range()
        print(Fore.GREEN + '[*] Перехват пакетов...\n')
        sniff(iface=iface_win(), filter='dst port 53', count=0, store=False, prn=packet_sniff)


if __name__ == "__main__":
    main()
